/* only for testing the functions */

#define USER_GMP_WRAP
#include "polynomial.cc"

// to check the stuff do:  g++ -O2 -g -Wall check_polynomial.cc -lgmp
// and for  dft-check do:  g++ -O2 -g -Wall -DUSE_DFT check_polynomial.cc -lgmp

using namespace polynomial;


bool do_check(const int maxk, const int anzPunkte, const mpz_t m)
{
  cout << endl << endl << "====================================" << endl;
  cout << " MULTIPOINT POLYNOM-EVAL SANITY " << maxk << ", " << anzPunkte << endl; 

  int i;

  mpz_t Polynom [maxk];
  mpz_t Punkte [anzPunkte];

  {
   const size_t bits = mpz_sizeinbase(m,2);

   gmp_randstate_t randstate;
   gmp_randinit_default(randstate);

   for (i=0; i<maxk; ++i) { mpz_init(Polynom[i]); mpz_rrandomb(Polynom[i],randstate,bits); mpz_mod(Polynom[i],Polynom[i],m); }
   for (i=0; i<anzPunkte; ++i) { mpz_init(Punkte[i]); mpz_rrandomb(Punkte[i],randstate,bits); mpz_mod(Punkte[i],Punkte[i],m); }

   gmp_randclear(randstate);
  }

  mpz_t Res_Horner [anzPunkte];
  for (i=0; i<anzPunkte; ++i) mpz_init(Res_Horner[i]);
 
  mpz_t Res_Multipoint [anzPunkte];
  for (i=0; i<anzPunkte; ++i) mpz_init(Res_Multipoint[i]);

  // testweise mit Hornerschema auswerten
  cout << "Starte Horner" << endl;
  time_t Start = time(NULL);
  for (i=0; i<anzPunkte; ++i) eval(Res_Horner[i],Polynom,maxk,Punkte[i],m);
  time_t Stop = time(NULL);
  cout << difftime(Stop,Start) << " Sekunden." << endl;

  // nun Multipoint-Auswertung
  cout << "Starte Multipoint" << endl;
  Start = time(NULL);
  multipoint_eval(Res_Multipoint,Polynom,maxk,Punkte,anzPunkte,m);
  Stop = time(NULL);
  cout << difftime(Stop,Start) << " Sekunden." << endl;

  cout << "berprfe Punkte fr Horner <-> Multipoint..." << endl;
  bool error = false;
  for (int i=0; i<anzPunkte; ++i)
   {
#ifdef DEBUG
     mpz_out_str(stdout,10,Punkte[i]); cout << " -> ";
     mpz_out_str(stdout,10,Res_Horner[i]); cout << " (Horner)  ";
     mpz_out_str(stdout,10,Res_Multipoint[i]); cout << " (Multipoint)" << endl;
#endif
     if (mpz_cmp(Res_Horner[i],Res_Multipoint[i]))
       {
	 error=true;
	 cerr << "Abweichung!! " << i << endl;
         mpz_out_str(stdout,10,Punkte[i]); cout << " -> ";
         mpz_out_str(stdout,10,Res_Horner[i]); cout << " (Horner)  ";
         mpz_out_str(stdout,10,Res_Multipoint[i]); cout << " (Multipoint)" << endl;
         break;
       }
   }
  if (error)
   {
     cout << "FAILED!" << endl;
     exit(1);
   }

  cout << "bereinstimmung festgestellt." << endl;
  cout << "PASSED." << endl;

  for (i=0; i<maxk; ++i) { mpz_clear(Polynom[i]); }
  for (i=0; i<anzPunkte; ++i) mpz_clear(Punkte[i]);
  for (i=0; i<anzPunkte; ++i) mpz_clear(Res_Horner[i]);
  for (i=0; i<anzPunkte; ++i) mpz_clear(Res_Multipoint[i]);
  return true;
}

bool performance_check(const int maxk, const int anzPunkte, const mpz_t m)
{
  cout << endl << endl << "====================================" << endl;
  cout << " MULTIPOINT POLYNOM-EVAL PERFORMANCE " << maxk << ", " << anzPunkte << endl; 

  int i;

  mpz_t Polynom [maxk];
  mpz_t Punkte [anzPunkte];

  {
   const size_t bits = mpz_sizeinbase(m,2);

   gmp_randstate_t randstate;
   gmp_randinit_default(randstate);

   for (i=0; i<maxk; ++i) { mpz_init(Polynom[i]); mpz_rrandomb(Polynom[i],randstate,bits); mpz_mod(Polynom[i],Polynom[i],m); }
   for (i=0; i<anzPunkte; ++i) { mpz_init(Punkte[i]); mpz_rrandomb(Punkte[i],randstate,bits); mpz_mod(Punkte[i],Punkte[i],m); }

   gmp_randclear(randstate);
  }


  mpz_t Res_Multipoint [anzPunkte];
  for (i=0; i<anzPunkte; ++i) mpz_init(Res_Multipoint[i]);

  // nun Multipoint-Auswertung
  cout << "Starte Multipoint" << endl;
  time_t Start = time(NULL);
  multipoint_eval(Res_Multipoint,Polynom,maxk,Punkte,anzPunkte,m);
  time_t Stop = time(NULL);
  cout << difftime(Stop,Start) << " Sekunden." << endl;

  for (i=0; i<maxk; ++i) { mpz_clear(Polynom[i]); }
  for (i=0; i<anzPunkte; ++i) mpz_clear(Punkte[i]);
  for (i=0; i<anzPunkte; ++i) mpz_clear(Res_Multipoint[i]);
  return true;
}


int main()
{
  cout << endl << endl
       << "===============================================" << endl;
  cout << "POLYNOMIAL-MULTIPOINT-EVAL doing some checks..." << endl;
  cout << "===============================================" << endl;

  mpz_t m;
  mpz_init(m);

  // 120-stellige Zahl generieren
  mpz_set_ui(m,10); mpz_pow_ui(m,m,120); mpz_add_ui(m,m,3);
//    mpz_set_ui(m,10); mpz_pow_ui(m,m,250); mpz_add_ui(m,m,3);

#if 0
  // torture check...
  for (int maxk=4; maxk<(1<<14); maxk*=2)
   {
     for (int i=3; i<maxk; ++i) do_check(maxk,i,m);
     for (int i=3; i<maxk+1; ++i) do_check(maxk+1,i,m);
   }
#endif


#if 1
  for (int maxk=1024; maxk<(1<<19); maxk*=2)
   {
     performance_check(maxk+1,maxk,m);
   }
#endif

  //performance_check(32769,32768,m);

#if 0
  // now a normal check
  for (int maxk=4; maxk<(1<<16); maxk*=2)
   {
     do_check(maxk,maxk-1,m);
     do_check(maxk+1,maxk,m);
   }

  mpz_set_str(m,"3923385745693995079670229419275984584311007321932374190635656246740175165573932140787529348954892963218868359081838772941945556717",10);
  for (int maxk=4; maxk<2048; maxk*=2)
   {
     do_check(maxk,maxk-1,m);
     do_check(maxk+1,maxk,m);
   }
#endif

  mpz_clear(m);
}
