#ifndef mpz_wrapper_dot_H_header
#define mpz_wrapper_dot_H_header


/*! @file
 * @brief
 * contains wrapped C++ stream I/O for mpz_t (multiple precision numbers)
 */

#include <iomanip>
#include <iostream>
#include <string>

//! contains wrapped I/O for mpz_t (multiple precision numbers)
namespace my_mpz_wrapper
{

using std::cerr;
using std::endl;

const int mpzbase_f = 32; // mpz-Ein/Ausgabeformat fr Fileoperationen
// remark:
// mpzbase_f=36: highest allowed base (refer gmp manual),
//               compatible with former versions (up to V2.95),
//               less space, but also computational less efficient.
// mpzbase_f=32: is a power of two (and therefore computational more efficient
//               since conversion routines can use shifts instead of expensive
//               divisions; refer gmp source).

// Fr lesbare Ein- und Ausgaben wird weiterhin Dezimalsystem verwendet  

inline std::ostream& operator << (std::ostream& ostr, const mpz_t value)
{ 
  // Ausgabefunktion fr mpz_t - Zahlen (um die hassenswerten expliziten Konvertierungen zu minimieren...)
  size_t size = mpz_sizeinbase(value,10); 
  char* str = new char [size+2];
  str[size]='\1'; // to suppress warnings about uninitialised value(s) issued by valgrind 
  mpz_get_str(str, 10, value);
  if (str[size])
   {
     if (str[size-1]=='\0') --size;
     else if (str[size+1]=='\0') ++size;
     else
      {
        MARK; exit(1);
      }
   }  
  ostr << str << " (" << size << ")";
  delete [] str;              
  return ostr;
}

inline std::istream& operator >> (std::istream& istr, mpz_t value)
{ 
  // Eingabefunktion fr mpz_t - Zahlen
  std::string s;
  istr >> s;
  if (mpz_set_str(value, s.c_str(), 10))
   {
     MARK;
     cerr << "invalid number read." << endl;
     cerr << "given: *" << s << "*" << endl;
     cerr << "mpz_t: *" << value << "*" << endl;
     istr.setstate(std::ios::failbit);
     exit(1);
   } else  istr >> s; // Sizeinbase berlesen
  return istr;
}

} // end of namespace my_mpz_wrapper

#endif /* of mpz_wrapper_dot_H_header */
